# -*- encoding: utf-8 -*-
# Author: Epix
import argparse
import json
# import ujson as json
import os
import sys
from pathlib import Path

from flask import Flask, send_file, jsonify


def start_server(size_tree_file_path, host, port, compression):
    app = Flask(__name__)
    base_dir = Path(__file__).absolute().parent / 'static' / 'dist' / 'webapp'
    app.root_path = str(base_dir)
    if compression:
        from flask_compress import Compress
        Compress(app)

    @app.route('/')
    def index():
        return send_file('index.html')

    @app.route('/<path:path>')
    def static_files(path):
        return send_file(path)

    @app.route('/size_tree.json')
    def size_tree():
        return send_file(size_tree_file_path, max_age=-1)

    @app.route('/info')
    def get_info():
        info = {
            'sep': os.path.sep
        }
        return jsonify(info)

    print('listening {}:{}'.format(host, port))
    app.run(host, port)


def scan_paths(root_paths, size_tree_file_path, args):
    all_size_tree = {}
    for root_path in root_paths:
        if root_path.startswith('s3://'):
            if root_path.endswith('/'):
                root_path = root_path[:-1]
            from .scan_s3 import scan_size_tree
            size_tree = scan_size_tree(root_path, args.endpoint_url)
        else:
            root_path = str(Path(root_path))
            if args.everything:
                from .scan_everything import scan_size_tree
                size_tree = scan_size_tree(root_path)
                if not size_tree:
                    from .scan_fs import scan_size_tree
                    size_tree = scan_size_tree(root_path, args.follow_links, args.follow_mounts)
            else:
                from .scan_fs import scan_size_tree
                size_tree = scan_size_tree(root_path, args.follow_links, args.follow_mounts)
        all_size_tree.update(size_tree)
    with open(size_tree_file_path, 'w') as f:
        json.dump(all_size_tree, f)
    print('scanning complete.')


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('paths', nargs='*',
                        help='path(s) to scan. '
                             'If multiple paths is provided, they will be show in root side by side. '
                             'S3 or compatible object storage service is supported by a "s3://" prefixed URI')
    parser.add_argument('--size-tree-path', '--size_tree_path', '-f', default='size_tree.json',
                        help='path to save scan result as a JSON file')
    parser.add_argument('--overwrite', '-o', action='store_true',
                        help='overwrite existed JSON file. default to False')
    parser.add_argument('--scan-only', '--scan_only', '-s', action='store_true',
                        help='scan and save JSON file but do not start web server. default to False')
    parser.add_argument('--host', '-H', default='127.0.0.1',
                        help='listening host of the web server')
    parser.add_argument('--port', '-p', default=8000, type=int,
                        help='listening port of the web server. default to 8000')
    parser.add_argument('--compression', '-c', action='store_true',
                        help='enable compression of web server. require flask_compress to operate. default to False')
    parser.add_argument('--endpoint-url', help='custom endpoint url, only affects S3')
    parser.add_argument('--follow-links', '--follow_links', action='store_true',
                        help='follow symlinks')
    parser.add_argument('--follow-mounts', '--follow_mounts', action='store_true',
                        help='follow mounts')
    parser.add_argument('--everything', action='store_true',
                        help='use Everything by voidtools to speedup scanning. The result will be absolute path. '
                             'Everything must be running and only x64 version is supported.')
    # parser.add_argument('--mlocate', nargs='?', const=True, default=False,
    #                     help='use mlocate to speedup scanning. you may specify location of mlocate.db file')
    args = parser.parse_args()
    root_paths = args.paths
    size_tree_file_path = os.path.abspath(args.size_tree_path)
    if os.path.exists(size_tree_file_path):
        if args.overwrite:
            scan_paths(root_paths, size_tree_file_path, args)
        else:
            print('{} exists. Skip scanning process.'.format(args.size_tree_path))
    else:
        if len(root_paths) == 0:
            print('nothing to scan and nothing to show. exiting.')
            return -1
        else:
            scan_paths(root_paths, size_tree_file_path, args)

    if not args.scan_only:
        start_server(size_tree_file_path=size_tree_file_path, host=args.host, port=args.port,
                     compression=args.compression)
    return 0


if __name__ == '__main__':
    sys.exit(main())



# -*- encoding: utf-8 -*-
# Author: Epix
import ctypes
import os
from pathlib import Path, PureWindowsPath

from tqdm import tqdm

# copied from https://www.voidtools.com/support/everything/sdk/python/

# defines
EVERYTHING_REQUEST_FILE_NAME = 0x00000001
EVERYTHING_REQUEST_PATH = 0x00000002
EVERYTHING_REQUEST_SIZE = 0x00000010
EVERYTHING_REQUEST_ATTRIBUTES = 0x00000100

# dll imports
try:
    everything_dll = ctypes.WinDLL(str(Path(__file__).parent.resolve() / "Everything64.dll"))
except:
    raise RuntimeError('Load DLL failed. Only Windows x64 is supported')
everything_dll.Everything_GetResultSize.argtypes = [ctypes.c_int, ctypes.POINTER(ctypes.c_ulonglong)]
everything_dll.Everything_GetResultFileNameW.argtypes = [ctypes.c_int]
everything_dll.Everything_GetResultFileNameW.restype = ctypes.c_wchar_p


def scan(top_path):
    prefix_len = len(top_path.rstrip(os.sep)) + 1
    if os.name == 'nt':
        if ':' in top_path:
            prefix_len -= 1  # backslash after drive letter and colon is missing under windows

    # setup search
    everything_dll.Everything_SetSearchW(f'"{top_path}"')
    everything_dll.Everything_SetRequestFlags(
        EVERYTHING_REQUEST_FILE_NAME | EVERYTHING_REQUEST_PATH | EVERYTHING_REQUEST_SIZE)

    # execute the query
    everything_dll.Everything_QueryW(1)

    # get the number of results
    num_results = everything_dll.Everything_GetNumResults()

    # create buffers
    filename = ctypes.create_unicode_buffer(260)
    file_size = ctypes.c_ulonglong(1)

    directory_size = 2 ** 64 - 1

    # show results
    for i in range(num_results):
        everything_dll.Everything_GetResultFullPathNameW(i, filename, 260)
        everything_dll.Everything_GetResultSize(i, file_size)
        if file_size.value == directory_size: # no way to judge if it's a directory
            continue
        yield ctypes.wstring_at(filename)[prefix_len:], file_size.value


def scan_size_tree(root_path):
    root_path = str(PureWindowsPath(Path(root_path).absolute()))
    size_tree = dict()
    for path, size in tqdm(scan(root_path), desc=root_path):
        *directories, filename = path.lstrip(os.sep).split(os.sep)
        directories.insert(0, root_path)
        current_size_tree = size_tree
        for directory in directories:
            if directory not in current_size_tree:
                current_size_tree[directory] = dict()
            current_size_tree = current_size_tree[directory]
        current_size_tree[filename] = size
    return size_tree



# -*- coding: utf-8 -*-
# Author: Epix
from functools import lru_cache

import boto3
from tqdm import tqdm


@lru_cache(maxsize=1)
def get_s3_client2(endpoint_url):
    return boto3.client('s3', endpoint_url=endpoint_url)


def walk_with_size(s3_url, endpoint_url):
    s3_url = s3_url + '/'
    bucket, prefix = s3_url[5:].split('/', 1)
    s3_client = get_s3_client2(endpoint_url)
    paginator = s3_client.get_paginator('list_objects_v2')
    response_iterator = paginator.paginate(Bucket=bucket, Prefix=prefix)
    for response in response_iterator:
        yield [(f['Key'], f['Size']) for f in response.get('Contents', ())]


def scan(top_path, endpoint_url):
    for files in walk_with_size(top_path, endpoint_url):
        for filename, size in files:
            yield filename, size


def scan_size_tree(root_path, endpoint_url):
    object_prefix_same = set()
    size_tree = dict()
    for path, size in tqdm(scan(root_path, endpoint_url), desc=root_path):
        *directories, filename = path.lstrip('/').split('/')
        directories.insert(0, root_path)
        current_size_tree = size_tree
        skip_flag = False
        for index, directory in enumerate(directories):
            if isinstance(current_size_tree, int):
                same_path = '/'.join(directories[:index])
                if same_path not in object_prefix_same:
                    object_prefix_same.add(same_path)
                    tqdm.write('s3 path: {} is both object and prefix'.format(same_path))
                skip_flag = True
                break
            if directory not in current_size_tree:
                current_size_tree[directory] = dict()
            current_size_tree = current_size_tree[directory]
        if skip_flag:
            continue
        if isinstance(current_size_tree, int):
            same_path = '/'.join(directories)
            if same_path not in object_prefix_same:
                object_prefix_same.add(same_path)
                tqdm.write('s3 path: {} is both object and prefix'.format(same_path))
            continue
        current_size_tree[filename] = size
    return size_tree



# -*- encoding: utf-8 -*-
# Author: Epix
import os

from tqdm import tqdm


def scan(top_path, follow_links, follow_mounts):
    prefix_len = len(top_path.rstrip(os.sep)) + 1
    if os.name == 'nt':
        if ':' in top_path:
            prefix_len -= 1  # backslash after drive letter and colon is missing under windows

    for root, dirs, filenames in os.walk(top_path, followlinks=follow_links):
        if not follow_mounts:
            dirs[:] = filter(lambda d: not os.path.ismount(os.path.join(root, d)), dirs)
        try:
            for filename in filenames:
                file_path = os.path.join(root, filename)
                yield file_path[prefix_len:], os.path.getsize(file_path)
        except (FileNotFoundError, PermissionError, OSError):
            pass


def scan_size_tree(root_path, follow_links, follow_mounts):
    size_tree = dict()
    for path, size in tqdm(scan(root_path, follow_links, follow_mounts), desc=root_path):
        *directories, filename = path.lstrip(os.sep).split(os.sep)
        directories.insert(0, root_path)
        current_size_tree = size_tree
        for directory in directories:
            if directory not in current_size_tree:
                current_size_tree[directory] = dict()
            current_size_tree = current_size_tree[directory]
        current_size_tree[filename] = size
    return size_tree



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import subprocess

import setuptools
from setuptools import setup
from disk_treemap import __version__


class BuildStatic(setuptools.Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        cwd = 'disk_treemap/static'
        subprocess.call(['npm', 'install'], cwd=cwd)
        subprocess.call(['npm', 'run', 'build', '--', '--prod'], cwd=cwd)


with open('README.md', 'r', encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='disk_treemap',
    version=__version__,
    author='Epix Zhang',
    packages=['disk_treemap'],
    description='Just another disk usage analyzer with treemap GUI.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/exzhawk/disk_treemap',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Framework :: Flask",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: System Administrators",
        "Programming Language :: JavaScript",
        "Topic :: Desktop Environment :: File Managers",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: System :: Filesystems",
        "Topic :: Utilities",
    ],
    python_requires='>=3.6',
    license='MIT',
    zip_safe=True,
    include_package_data=True,
    cmdclass={
        'build_static': BuildStatic,
    },
    install_requires=[
        'flask>=2.0',
        'tqdm',
        'flask-compress',
        'boto3',
    ],
    entry_points={
        'console_scripts': [
            'disk-treemap=disk_treemap.main:main',
        ]
    }
)
